use anyhow::{Error, Result};
use once_cell::sync::OnceCell;
use std::sync::{Arc, RwLock, RwLockReadGuard, RwLockWriteGuard};

mod hardware;
mod swu;

pub use hardware::HardwareConfig;
pub use swu::SwuConfig;

static CONFIG: OnceCell<Config> = OnceCell::new();

// Outwards-facing struct: wrap everything in an Arc/RwLock
// so it's safe to clone/update.
#[derive(Clone, Debug)]
pub struct Config {
    #[allow(unused)]
    inner: Arc<RwLock<ConfigInner>>,
}

// (clone is only used for default config)
#[derive(Clone, Debug)]
pub struct ConfigInner {
    pub hardware: HardwareConfig,
    pub swu: SwuConfig,
}

impl Config {
    async fn do_load() -> ConfigInner {
        let hardware = HardwareConfig::load().await;
        let swu = SwuConfig::load().await;
        ConfigInner { hardware, swu }
    }

    pub async fn load() -> Result<()> {
        let inner = Config::do_load().await;
        CONFIG
            .set(Config {
                inner: RwLock::new(inner).into(),
            })
            .expect("Load should only be called once");
        Ok(())
    }

    pub fn get() -> &'static Config {
        CONFIG.get().expect("Config is not initialized")
    }

    // utility functions to get inner lock directly
    pub fn read(&self) -> Result<RwLockReadGuard<ConfigInner>> {
        // lock's PoisonError isn't convertible to anyhow Error, map it
        self.inner
            .read()
            .map_err(|_| Error::msg("Could not get config read lock"))
    }
    pub fn _write(&self) -> Result<RwLockWriteGuard<ConfigInner>> {
        self.inner
            .write()
            .map_err(|_| Error::msg("Could not get config read lock"))
    }
}
