use anyhow::{anyhow, Context, Error, Result};
use ipnet::Ipv4Net;
use std::collections::HashMap;
use std::fs;
use std::path::Path;
use std::process::Output;
use tokio::process::Command;

mod config;
pub mod container;
pub mod html_template;
mod iptables;
pub mod link_properties;
pub mod networkmanager;
pub mod side_menu;

pub use config::Config;
pub use iptables::{IptablesInfo, NatInfo, PortForwardingInfo};

#[cfg(debug_assertions)]
pub const SCRIPTS_DIR: &str = "./scripts";

#[cfg(not(debug_assertions))]
pub const SCRIPTS_DIR: &str = "/usr/libexec/abos-web";

pub const WWAN_CON_NAME: &str = "gsm-ttyCommModem";
pub const WLAN_CON_NAME: &str = "abos_web_wlan";
pub const AP_BRIDGE_INTERFACE: &str = "br_ap";
pub const AP_BRIDGE_NAME: &str = "abos_web_br_ap";
pub const VERSION: &str = env!("CARGO_PKG_VERSION");
pub const OPENVPN_CONF_DIR: &str = "/etc/atmark/abos_web/openvpn";
pub const OPENVPN_CONTAINER: &str = "abos_web_openvpn";

pub fn ap_interface() -> Option<&'static str> {
    if Path::new("/sys/class/net/uap0").is_dir() {
        Some("uap0")
    } else if Path::new("/sys/class/net/wlan0").is_dir() {
        Some("wlan0")
    } else {
        None
    }
}

pub fn wlan_interface() -> Option<&'static str> {
    if Path::new("/sys/class/net/mlan0").is_dir() {
        Some("mlan0")
    } else if Path::new("/sys/class/net/wlan0").is_dir() {
        Some("wlan0")
    } else {
        None
    }
}

pub fn wwan_interface() -> Option<&'static str> {
    if Path::new("/sys/class/net/usb0").is_dir() {
        Some("usb0")
    } else if Path::new("/sys/class/net/ppp0").is_dir() {
        Some("ppp0")
    } else {
        None
    }
}

pub fn wwan_module() -> Option<&'static str> {
    if Path::new("/sys/devices/platform/ec25-reset").is_dir() {
        Some("ec25")
    } else if Path::new("/sys/devices/platform/els31-reset").is_dir() {
        Some("els31")
    } else if Path::new("/sys/devices/platform/ems31-reset").is_dir() {
        Some("ems31")
    } else {
        None
    }
}

pub fn network_interfaces() -> Vec<String> {
    let mut v = Vec::new();
    if let Ok(entries) = fs::read_dir("/sys/class/net") {
        for entry in entries.flatten() {
            v.push(entry.file_name().to_str().unwrap_or_default().to_string());
        }
    }
    v
}

pub fn exists_eth1() -> bool {
    Path::new("/sys/class/net/eth1").is_dir()
}

pub fn exists_br_ap() -> bool {
    Path::new("/sys/class/net/br_ap").is_dir()
}

pub fn doas() -> &'static str {
    if cfg!(debug_assertions) && Path::new("./utils/ssh_run.sh").is_file() {
        "./utils/ssh_run.sh"
    } else if Path::new("/usr/bin/doas").is_file() {
        "doas"
    } else {
        ""
    }
}

// take script_dir / doas as argument for tests
fn _command<S>(args: &[S], script_dir: Option<&str>, doas: &str) -> Result<Command>
where
    S: AsRef<std::ffi::OsStr>,
{
    if args.is_empty() {
        return Err(anyhow!("no arg"));
    }

    let mut alloc = String::new();
    let arg0 = match script_dir {
        None => args[0].as_ref(),
        Some(dir) => {
            alloc.push_str(dir);
            alloc.push('/');
            alloc.push_str(
                args[0]
                    .as_ref()
                    .to_str()
                    .context("Could not cast OsStr back to str")?,
            );
            alloc.as_ref()
        }
    };

    if doas.is_empty() {
        let mut command = Command::new(arg0);
        command.args(&args[1..]);
        Ok(command)
    } else {
        let mut command = Command::new(doas);
        command.arg(arg0);
        command.args(&args[1..]);
        Ok(command)
    }
}

pub struct CommandScript(pub bool);
pub struct CommandAsRoot(pub bool);

pub fn command<S>(args: &[S], script: CommandScript, as_root: CommandAsRoot) -> Result<Command>
where
    S: AsRef<std::ffi::OsStr>,
{
    let scripts_dir = if script.0 { Some(SCRIPTS_DIR) } else { None };
    let doas = if as_root.0 { doas() } else { "" };

    _command(args, scripts_dir, doas)
}

pub async fn exec_command<S>(args: &[S]) -> Result<Output>
where
    S: AsRef<std::ffi::OsStr> + std::fmt::Debug,
{
    let output = command(args, CommandScript(true), CommandAsRoot(true))?
        .output()
        .await?;
    if output.status.success() {
        Ok(output)
    } else {
        Err(Error::msg(format!(
            "Command {:?} failed with {}:\nstderr:\n{}\nstdout:\n{}",
            args,
            output.status,
            String::from_utf8_lossy(&output.stderr),
            String::from_utf8_lossy(&output.stdout)
        )))
    }
}

pub async fn current_address<'a>(
    dev_name: &str,
    interface: &str,
) -> Result<HashMap<&'a str, String>, &'static str> {
    let args = &["get_address.sh", dev_name, interface];
    let output = match exec_command(args).await {
        Ok(o) => o,
        Err(_) => return Err("Could not get current address."),
    };
    let stdout = String::from_utf8_lossy(&output.stdout);
    // addrs[0]: ip addr with netmask e.g. 192.168.1.1/24
    // addrs[1]: gateway addr
    // addrs[2]: DNS addr
    // addrs[3]: MAC addr
    let addrs = stdout
        .split(';')
        .collect::<Vec<&str>>()
        .iter()
        .map(|s| s.trim().to_string())
        .collect::<Vec<String>>();

    let na = "--".to_string();
    let mut map = HashMap::new();

    map.insert("current_ip_addr", na.clone());
    map.insert("current_netmask", na.clone());
    map.insert("current_gw_addr", na.clone());
    map.insert("current_dns_addr", na.clone());
    map.insert("mac_addr", na.clone());
    map.insert("interface", na);

    if !addrs[0].is_empty() {
        let net: Ipv4Net = match addrs[0].parse() {
            Ok(n) => n,
            Err(_) => return Err("Invalid IP Address."),
        };
        map.insert("current_ip_addr", net.addr().to_string());
        map.insert("current_netmask", net.netmask().to_string());
    }
    if !addrs[1].is_empty() {
        map.insert("current_gw_addr", addrs[1].to_string());
    }
    if !addrs[2].is_empty() {
        map.insert("current_dns_addr", addrs[2].to_string());
    }
    if !addrs[3].is_empty() {
        map.insert("mac_addr", addrs[3].to_string());
    }
    map.insert("interface", interface.to_string());

    Ok(map)
}

pub async fn ap_info() -> Result<HashMap<&'static str, String>, &'static str> {
    let args = &["ap_info.sh"];
    let output = match exec_command(args).await {
        Ok(o) => o,
        Err(_) => return Err("Could not get current address."),
    };
    let stdout = String::from_utf8_lossy(&output.stdout).trim().to_string();
    /*
        ap_info[0]: bridge addr with netmask e.g. 192.168.1.1/24
        ap_info[1]: hw_mode a:5GHz g:2.4GHz
        ap_info[2]: channel
        ap_info[3]: ssid
    */
    let ap_info = stdout
        .split(',')
        .collect::<Vec<_>>()
        .iter()
        .map(|s| s.trim().to_string())
        .collect::<Vec<String>>();

    let na = "--".to_string();
    let mut map = HashMap::new();

    map.insert("bridge_addr", na.clone());
    map.insert("netmask", na.clone());
    map.insert("hw_mode", na.clone());
    map.insert("channel", na.clone());
    map.insert("ssid", na.clone());
    map.insert("interface", na);

    if !stdout.is_empty() && !ap_info.is_empty() {
        if !ap_info[0].is_empty() {
            let net: Ipv4Net = match ap_info[0].parse() {
                Ok(n) => n,
                Err(_) => return Err("Invalid IP Address."),
            };
            map.insert("bridge_addr", net.addr().to_string());
            map.insert("netmask", net.netmask().to_string());
        }
        if !ap_info[1].is_empty() {
            map.insert("hw_mode", ap_info[1].to_string());
        }
        if !ap_info[2].is_empty() {
            map.insert("channel", ap_info[2].clone());
        }
        if !ap_info[3].is_empty() {
            map.insert("ssid", ap_info[3].to_string());
        }
        map.insert("interface", AP_BRIDGE_INTERFACE.to_string());
    }

    Ok(map)
}

pub async fn dhcp_info(interface: &str) -> Result<HashMap<&'static str, String>, &'static str> {
    let args = &["dhcp_info.sh", interface];
    let output = match exec_command(args).await {
        Ok(o) => o,
        Err(_) => return Err("Could not get current address."),
    };
    let stdout = String::from_utf8_lossy(&output.stdout).trim().to_string();
    /*
        ap_info[0]: ip addr with netmask e.g. 192.168.1.1/24
        ap_info[1]: DHCP start addr
        ap_info[2]: DHCP end addr
        ap_info[3]: DHCP lease time
    */
    let dhcp_info = stdout
        .split(',')
        .collect::<Vec<_>>()
        .iter()
        .map(|s| s.trim().to_string())
        .collect::<Vec<String>>();

    let mut map = HashMap::new();
    if !stdout.is_empty() && !dhcp_info.is_empty() {
        if !dhcp_info[0].is_empty() {
            let net: Ipv4Net = match dhcp_info[0].parse() {
                Ok(n) => n,
                Err(_) => return Err("Invalid IP Address."),
            };
            map.insert("ip_addr", net.addr().to_string());
            map.insert("netmask", net.netmask().to_string());
        } else {
            map.insert("ip_addr", "--".to_string());
            map.insert("netmask", "--".to_string());
        }
        if !dhcp_info[1].is_empty() {
            map.insert("start_addr", dhcp_info[1].to_string());
        }
        if !dhcp_info[2].is_empty() {
            map.insert("end_addr", dhcp_info[2].to_string());
        }
        if !dhcp_info[3].is_empty() {
            map.insert("lease_time", dhcp_info[3].to_string());
        }
        map.insert("interface", interface.to_string());
    }

    Ok(map)
}

pub async fn wwan_info(con_name: &str) -> Result<HashMap<&'static str, String>, &'static str> {
    let args = &["wwan_info.sh", con_name];
    let output = match exec_command(args).await {
        Ok(o) => o,
        Err(_) => return Err("Could not get wwan info."),
    };
    let stdout = String::from_utf8_lossy(&output.stdout).trim().to_string();
    /*
        wwan_info[0]: APN
        wwan_info[1]: USER NAME
        wwan_info[2]: MCC/MNC
        wwan_info[3]: AUTH TYPE
        wwan_info[4]: IMSI
        wwan_info[5]: SIGNALQ
    */
    let wwan_info = stdout
        .split(';')
        .collect::<Vec<_>>()
        .iter()
        .map(|s| s.trim().to_string())
        .collect::<Vec<String>>();

    let keys = vec!["apn", "user_name", "mccmnc", "auth", "imsi", "signalq"];
    let mut map = HashMap::new();
    for key in keys.iter() {
        map.insert(*key, "--".to_string());
    }
    for (i, info) in wwan_info.iter().enumerate() {
        if !info.is_empty() {
            map.insert(keys[i], info.to_string());
        }
    }
    Ok(map)
}

#[cfg(test)]
mod tests {
    use crate::common::_command;
    use anyhow::Result;
    use std::process::Command;

    fn check_command(cmd: &Command, arg0: &str, args: &[&str]) {
        assert_eq!(cmd.get_program(), arg0);
        assert_eq!(cmd.get_args().count(), args.len());
        cmd.get_args()
            .zip(args)
            .for_each(|(cmd_arg, check_arg)| assert_eq!(&cmd_arg.to_str().unwrap(), check_arg));
    }

    #[test]
    fn test_command() -> Result<()> {
        assert!(_command(&Vec::<String>::new(), None, "").is_err());

        // single argument
        let args = &["ls"];
        let cmd = _command(args, None, "")?;
        check_command(cmd.as_std(), "ls", &[]);
        let cmd = _command(&args.map(|s| s.to_string()), None, "")?;
        check_command(cmd.as_std(), "ls", &[]);

        let cmd = _command(args, None, "doas")?;
        check_command(cmd.as_std(), "doas", &["ls"]);
        let cmd = _command(&args.map(|s| s.to_string()), None, "doas")?;
        check_command(cmd.as_std(), "doas", &["ls"]);

        let cmd = _command(args, Some("./scripts"), "")?;
        check_command(cmd.as_std(), "./scripts/ls", &[]);
        let cmd = _command(&args.map(|s| s.to_string()), Some("./scripts"), "")?;
        check_command(cmd.as_std(), "./scripts/ls", &[]);

        let cmd = _command(args, Some("./scripts"), "doas")?;
        check_command(cmd.as_std(), "doas", &["./scripts/ls"]);
        let cmd = _command(&args.map(|s| s.to_string()), Some("./scripts"), "doas")?;
        check_command(cmd.as_std(), "doas", &["./scripts/ls"]);

        // some argument
        let args = &["ls", "-l"];
        let cmd = _command(args, None, "")?;
        check_command(cmd.as_std(), "ls", &["-l"]);
        let cmd = _command(&args.map(|s| s.to_string()), None, "")?;
        check_command(cmd.as_std(), "ls", &["-l"]);

        let cmd = _command(args, None, "doas")?;
        check_command(cmd.as_std(), "doas", &["ls", "-l"]);
        let cmd = _command(&args.map(|s| s.to_string()), None, "doas")?;
        check_command(cmd.as_std(), "doas", &["ls", "-l"]);

        let cmd = _command(args, Some("./scripts"), "")?;
        check_command(cmd.as_std(), "./scripts/ls", &["-l"]);
        let cmd = _command(&args.map(|s| s.to_string()), Some("./scripts"), "")?;
        check_command(cmd.as_std(), "./scripts/ls", &["-l"]);

        let cmd = _command(args, Some("./scripts"), "doas")?;
        check_command(cmd.as_std(), "doas", &["./scripts/ls", "-l"]);
        let cmd = _command(&args.map(|s| s.to_string()), Some("./scripts"), "doas")?;
        check_command(cmd.as_std(), "doas", &["./scripts/ls", "-l"]);

        Ok(())
    }
}
