use anyhow::Context;
use askama::Template;
use axum::{
    extract::Form,
    response::{IntoResponse, Redirect},
    routing::{get, post},
    Router,
};
use axum_sessions::extractors::ReadableSession;
use serde::Deserialize;
use std::collections::HashMap;

use crate::common;
use crate::common::{
    html_template::HtmlTemplate,
    networkmanager::{nmconmap, NMCon},
};
use crate::error::PageResult;

#[derive(Deserialize)]
struct DhcpParam {
    interface: String,
    start_addr: String,
    end_addr: String,
    lease_time: String,
}

#[derive(Deserialize)]
struct DhcpDelParam {
    interface: String,
}

struct DhcpCandidate {
    interface: String,
    ip: String,
}

#[derive(Template)]
#[template(path = "../src/dhcp/templates/dhcp.html")]
struct DhcpTemplate {
    candidates: Vec<DhcpCandidate>,
    dhcp_infos: Vec<HashMap<&'static str, String>>,
}

pub trait DhcpRoutes {
    fn dhcp_routes(self) -> Self;
}

impl DhcpRoutes for Router {
    fn dhcp_routes(self) -> Self {
        self.route("/dhcp", get(dhcp))
            .route("/dhcp_setup", post(dhcp_setup))
            .route("/dhcp_delete", post(dhcp_delete))
    }
}

async fn dhcp(session: ReadableSession) -> PageResult {
    if !cfg!(debug_assertions) && !session.get::<bool>("login").unwrap_or(false) {
        return Ok(Redirect::to("/").into_response());
    }

    let cons = nmconmap().await?;
    let mut candidates = Vec::<DhcpCandidate>::new();
    let mut dhcp_infos = Vec::<HashMap<&'static str, String>>::new();
    for uuid in cons.keys() {
        // cons values don't have enough info, will add a helper to enrich later
        // but for now cycle manually...
        // ignore errors, convert to if let Some() else continue after rust upgrade...
        if let Ok(con) = NMCon::from_id(uuid).await {
            // ignore interfaces not in manual mode
            if con
                .prop("ipv4.method")
                .map(|v| v != "manual")
                .unwrap_or(true)
            {
                continue;
            }
            if con.device.is_none() {
                continue;
            }
            let interface = con
                .device
                .as_ref()
                .context("no interface after checking none")?
                .as_str();
            // ignore podman interfaces
            if interface.starts_with("podman") {
                continue;
            }
            let dhcp_info = common::dhcp_info(interface).await?;
            if dhcp_info.is_empty() {
                // candidate
                if let Some(ip) = con.prop("IP4.ADDRESS[1]") {
                    candidates.push(DhcpCandidate {
                        interface: interface.to_string(),
                        ip: ip.to_string(),
                    });
                }
            } else {
                dhcp_infos.push(dhcp_info);
            }
        }
    }

    candidates.sort_unstable_by(|a, b| a.interface.cmp(&b.interface));
    dhcp_infos.sort_unstable_by(|a, b| a["interface"].cmp(&b["interface"]));

    let template = DhcpTemplate {
        candidates,
        dhcp_infos,
    };
    Ok(HtmlTemplate::new("DHCPサーバー設定", template).into_response())
}

async fn dhcp_setup(session: ReadableSession, Form(dhcp_param): Form<DhcpParam>) -> PageResult {
    if !cfg!(debug_assertions) && !session.get::<bool>("login").unwrap_or(false) {
        return Ok(Redirect::to("/").into_response());
    }

    let args = &[
        "dhcp_setup.sh",
        &dhcp_param.start_addr,
        &dhcp_param.end_addr,
        &dhcp_param.lease_time,
        &dhcp_param.interface,
    ];
    common::exec_command(args).await?;
    Ok(Redirect::to("/dhcp").into_response())
}

async fn dhcp_delete(
    session: ReadableSession,
    Form(dhcp_del_param): Form<DhcpDelParam>,
) -> PageResult {
    if !cfg!(debug_assertions) && !session.get::<bool>("login").unwrap_or(false) {
        return Ok(Redirect::to("/").into_response());
    }

    let args = &["dhcp_delete.sh", &dhcp_del_param.interface];
    common::exec_command(args).await?;
    Ok(Redirect::to("/dhcp").into_response())
}
