use anyhow::Result;
use askama::Template;
use axum::{
    extract::Form,
    response::{IntoResponse, Redirect},
    routing::{get, post},
    Router,
};
use axum_sessions::extractors::WritableSession;
use once_cell::sync::OnceCell;
use pwhash::sha512_crypt;
use serde::Deserialize;
use std::time::Duration;
use tokio::sync::Mutex;
use tokio::time::Instant;

use crate::common;
use crate::common::html_template::RawTemplate;
use crate::error::PageResult;

static PASSWORD_THROTTLE: OnceCell<Mutex<()>> = OnceCell::new();

#[derive(Deserialize, Debug, Clone)]
struct User {
    password: String,
}

#[derive(Template)]
#[template(path = "../src/login/templates/login.html")]
struct LoginTemplate {
    login_error: bool,
    version: &'static str,
}

#[derive(Deserialize, Debug, Clone)]
struct RegisterPasswordParam {
    password: String,
    password_confirm: String,
}

#[derive(Template)]
#[template(path = "../src/login/templates/new_password.html")]
struct NewPasswordTemplate {
    confirm_error: bool,
    first_time: bool,
    version: &'static str,
}

#[derive(Template)]
#[template(path = "../src/login/templates/success.html")]
struct SuccessTemplate {
    version: &'static str,
}

pub trait LoginRoutes {
    fn login_routes(self) -> Self;
}

impl LoginRoutes for Router {
    fn login_routes(self) -> Self {
        self.route("/", get(index))
            .route("/login", post(login))
            .route("/logout", get(logout))
            .route("/new_password", get(new_password))
            .route("/register_password", post(register_password))
            .route("/success", get(success))
    }
}

pub fn password_throttle_init() {
    PASSWORD_THROTTLE
        .set(Mutex::new(()))
        .expect("init should only be called once");
}

pub async fn password_hash() -> Result<String> {
    if cfg!(debug_assertions) {
        return Ok("!".to_string());
    };
    let args = vec!["get_pw_hash.sh".to_string()];
    let output = common::exec_command(&args).await?;
    Ok(String::from_utf8_lossy(&output.stdout).replace('\n', ""))
}

async fn index(mut session: WritableSession) -> PageResult {
    if session.get::<bool>("login").unwrap_or(false) {
        return Ok(Redirect::to("/top").into_response());
    }
    if !cfg!(debug_assertions) {
        let hash = password_hash().await?;
        if hash == "!" {
            return Ok(Redirect::to("/new_password").into_response());
        }
    }

    let login_error = session.get::<bool>("login_error").unwrap_or(false);
    session.remove("login_error");
    let template = LoginTemplate {
        login_error,
        version: common::VERSION,
    };
    Ok(RawTemplate(template).into_response())
}

async fn login(mut session: WritableSession, Form(user): Form<User>) -> PageResult {
    let password = user.password;
    if cfg!(debug_assertions) {
        if password.is_empty() {
            return Ok(Redirect::to("/new_password").into_response());
        }
        if session.insert("login", true).is_err() {
            return Ok(Redirect::to("/").into_response());
        }
        Ok(Redirect::to("/top").into_response())
    } else {
        let _lock = PASSWORD_THROTTLE
            .get()
            .expect("password throttle should be initialized")
            .lock()
            .await;
        let start_time = Instant::now();
        let hash = password_hash().await?;
        if sha512_crypt::verify(password, &hash) {
            if session.insert("login", true).is_err() {
                return Ok(Redirect::to("/").into_response());
            }
            Ok(Redirect::to("/top").into_response())
        } else {
            tokio::time::sleep_until(start_time + Duration::from_secs(2)).await;
            let _ = session.insert("login_error", true);
            Ok(Redirect::to("/").into_response())
        }
    }
}

async fn logout(mut session: WritableSession) -> Redirect {
    session.destroy();
    Redirect::to("/")
}

async fn new_password(session: WritableSession) -> PageResult {
    let hash = password_hash().await?;
    if !session.get::<bool>("login").unwrap_or(false) && hash != "!" {
        return Ok(Redirect::to("/").into_response());
    }
    let confirm_error = session.get::<bool>("confirm_error").unwrap_or(false);
    let template = NewPasswordTemplate {
        confirm_error,
        first_time: hash == "!",
        version: common::VERSION,
    };
    Ok(RawTemplate(template).into_response())
}

async fn register_password(
    mut session: WritableSession,
    Form(password_param): Form<RegisterPasswordParam>,
) -> PageResult {
    if !session.get::<bool>("login").unwrap_or(false) && password_hash().await? != "!" {
        return Ok(Redirect::to("/").into_response());
    }

    if password_param.password.is_empty() || password_param.password_confirm.is_empty() {
        return Ok(Redirect::to("/new_password").into_response());
    }
    if password_param.password == password_param.password_confirm {
        if cfg!(debug_assertions) {
            Ok(Redirect::to("/success").into_response())
        } else {
            let args = &["register_password.sh", &password_param.password];
            match common::exec_command(args).await {
                Ok(o) => o,
                Err(_) => return Ok(Redirect::to("/").into_response()),
            };
            Ok(Redirect::to("/success").into_response())
        }
    } else {
        match session.insert("confirm_error", true) {
            Ok(_) => {}
            Err(_) => return Ok(Redirect::to("/").into_response()),
        }
        Ok(Redirect::to("/new_password").into_response())
    }
}

async fn success(mut session: WritableSession) -> impl IntoResponse {
    session.destroy();
    let template = SuccessTemplate {
        version: common::VERSION,
    };
    RawTemplate(template).into_response()
}
