use anyhow::{Context, Result};
use askama::Template;
use async_stream::try_stream;
use axum::{
    body::StreamBody,
    extract::{DefaultBodyLimit, Multipart},
    http::header::CONTENT_TYPE,
    response::{IntoResponse, Redirect},
    routing::{get, post},
    Router,
};
use axum_sessions::extractors::ReadableSession;
use futures::Stream;
use std::process::Stdio;
use tokio::io::{AsyncBufReadExt, AsyncWriteExt, BufReader};
use tokio::process::Child;
use tokio::sync::mpsc;
use tracing::{info, trace, warn};

use crate::common::{
    command, exec_command, html_template::HtmlTemplate, CommandAsRoot, CommandScript, Config,
};
use crate::error::{ErrorString, ErrorStringResult, PageResult};

mod version;
pub use version::{get_swu_versions, SwuVersion};

#[derive(Template)]
#[template(path = "../src/swu/templates/swu.html")]
struct SwuTemplate {
    initial_setup_installed: bool,
    swupdate_watch_url: String,
    swu_versions: Vec<SwuVersion>,
}

pub trait SwuRoutes {
    fn swu_routes(self) -> Self;
}

impl SwuRoutes for Router {
    fn swu_routes(self) -> Self {
        self.route("/swu", get(swu))
            .route("/swu_install_url", post(swu_install_url))
            // axum's default body limit is 2MB: we disable it just to
            // allow sending big SWUs, but restrict back other endpoints
            // lower in the stack
            .layer(DefaultBodyLimit::max(2048 * 1024))
            .route("/swu_install_file", post(swu_install_file))
            .layer(DefaultBodyLimit::disable())
            .route("/swu_latest_log", post(swu_latest_log))
    }
}

async fn swu(session: ReadableSession) -> PageResult {
    if !cfg!(debug_assertions) && !session.get::<bool>("login").unwrap_or(false) {
        return Ok(Redirect::to("/").into_response());
    }

    let swu_versions = get_swu_versions().await.unwrap_or_default();
    // note: cannot use await after getting this lock
    let swu_config = &Config::get().read()?.swu;
    let template = SwuTemplate {
        initial_setup_installed: swu_config.initial_setup_installed,
        swupdate_watch_url: swu_config.watch_urls.join(" "),
        swu_versions,
    };
    Ok(HtmlTemplate::new("SWU インストール", template).into_response())
}

async fn swu_install_url(session: ReadableSession, mut multipart: Multipart) -> impl IntoResponse {
    if !cfg!(debug_assertions) && !session.get::<bool>("login").unwrap_or(false) {
        // spell out the Error type once manually
        return Err(ErrorString {
            message: "not logged in".to_string(),
        });
    }
    let Ok(Some(url_data)) = multipart.next_field().await else {
        return Err("No multipart data".into());
    };
    let Some("url") = url_data.name() else {
        return Err("Multipart field name should be 'url'".into());
    };
    let Ok(url) = url_data.text().await else {
        return Err("Could not get 'url' text".into());
    };

    info!("installing from url {}", url);
    let mut args = vec!["url".to_string()];
    url.trim().split(' ').for_each(|u| {
        if !u.is_empty() {
            args.push(u.to_string())
        }
    });
    let chan = swupdate_install(args, None).await?;
    Ok((
        [(CONTENT_TYPE, "text/plain")],
        StreamBody::new(forward_output(chan).await),
    ))
}

async fn swu_latest_log(session: ReadableSession) -> ErrorStringResult {
    if !cfg!(debug_assertions) && !session.get::<bool>("login").unwrap_or(false) {
        // spell out the Error type once manually
        Err("not logged in")?;
    }
    Ok(exec_command(&["get_latest_swupdate_log.sh"])
        .await
        .map_err(|_| "ログを取得できませんでした")?
        .stdout
        .into_response())
}

async fn swu_install_file(session: ReadableSession, multipart: Multipart) -> impl IntoResponse {
    if !cfg!(debug_assertions) && !session.get::<bool>("login").unwrap_or(false) {
        // spell out the Error type once manually
        return Err(ErrorString {
            message: "not logged in".to_string(),
        });
    }

    let args = vec!["stdin".to_string()];
    let chan = swupdate_install(args, Some(multipart)).await?;
    Ok((
        [(CONTENT_TYPE, "text/plain")],
        StreamBody::new(forward_output(chan).await),
    ))
}

async fn swupdate_install(
    args: Vec<String>,
    input_data: Option<Multipart>,
) -> Result<mpsc::Receiver<Vec<u8>>> {
    let (tx, rx) = mpsc::channel(20);
    let command = command(&["swupdate.sh"], CommandScript(true), CommandAsRoot(true))?
        .args(args)
        .stdin(if input_data.is_none() {
            Stdio::null()
        } else {
            Stdio::piped()
        })
        .stdout(Stdio::piped())
        .stderr(Stdio::piped())
        .spawn()
        .context("Could not start swupdate")?;
    tokio::spawn(async move {
        if let Err(e) = swupdate_handle_io(&tx, input_data, command).await {
            info!("Swupdate install failed: {:?}", e);
            let _ = tx.send(e.to_string().into_bytes()).await;
        }
    });
    Ok(rx)
}
/// start update, messages will be obtained in given channel
async fn swupdate_handle_io(
    tx: &mpsc::Sender<Vec<u8>>,
    input_data: Option<Multipart>,
    mut command: Child,
) -> Result<()> {
    let mut stdout = BufReader::new(
        command
            .stdout
            .take()
            .context("swupdate command had no stdout")?,
    );
    let mut stdout_buf = vec![];
    let mut stdout_open = true;
    let mut stderr = BufReader::new(
        command
            .stderr
            .take()
            .context("swupdate command had no stderr")?,
    );
    let mut stderr_buf = vec![];
    let mut stderr_open = true;

    if let Some(mut input_multipart) = input_data {
        let mut stdin = command
            .stdin
            .take()
            .context("swupdate command had no stdin")?;
        let tx2 = tx.clone();
        tokio::spawn(async move {
            if let Ok(Some(mut input_stream)) = input_multipart.next_field().await {
                // check data.name()? == file ? content_type = binary?
                // print file_name to display what we're installing?
                // Note we can't unpack multipart out because the field has a mutable
                // reference to the multipart, and we cannot move {multipart, field}
                // together in the closure...
                // XXX ^ can now be done through channel
                loop {
                    match input_stream.chunk().await {
                        Ok(Some(chunk)) => {
                            if let Err(e) = stdin.write_all(&chunk).await {
                                warn!("failed to write stream for swupdate {:?}", e);
                                let _ = tx2.send(e.to_string().into_bytes()).await;
                                break;
                            }
                        }
                        Ok(None) => break, // end of stream
                        Err(e) => {
                            warn!("failed to read stream for swupdate {:?}", e);
                            let _ = tx2.send(e.to_string().into_bytes()).await;
                            break;
                        }
                    }
                }
            }
        });
    }

    while stdout_open || stderr_open {
        tokio::select! {
            n = stdout.read_until(b'\n', &mut stdout_buf), if stdout_open => {
                if n.unwrap_or(0) == 0 && stdout_buf.is_empty() {
                    stdout_open = false;
                } else {
                    trace!("{}", String::from_utf8_lossy(&stdout_buf));
                    // ignore errors: we continue update anyway
                    let _ = tx.send(stdout_buf).await;
                    stdout_buf = vec![];
                }
            },
            n = stderr.read_until(b'\n', &mut stderr_buf), if stderr_open => {
                if n.unwrap_or(0) == 0 && stderr_buf.is_empty() {
                    stderr_open = false;
                } else {
                    trace!("{}", String::from_utf8_lossy(&stderr_buf));
                    let _ = tx.send(stderr_buf).await;
                    stderr_buf = vec![];
                }
            },
        }
    }

    let _ = command.wait().await.context("waiting failed")?;
    // ignore failure as it exits with non-zero on reboot (self-kill)
    let _ = tx.send("swupdate exited.".as_bytes().to_vec()).await;
    Ok(())
}

async fn forward_output(mut rx: mpsc::Receiver<Vec<u8>>) -> impl Stream<Item = Result<Vec<u8>>> {
    try_stream! {
        while let Some(line) = rx.recv().await {
            yield line
        }
    }
}
