use anyhow::{anyhow, Context, Result};
use askama::Template;
use async_stream::try_stream;
use axum::{
    body::Bytes,
    body::StreamBody,
    http::header::CONTENT_TYPE,
    response::{IntoResponse, Redirect},
    routing::{get, post},
    Router,
};
use axum_typed_multipart::{FieldData, TryFromMultipart, TypedMultipart};

use axum_sessions::extractors::ReadableSession;
use futures::Stream;
use std::collections::HashMap;
use std::process::Stdio;

use crate::common;
use crate::common::{command, html_template::HtmlTemplate, CommandAsRoot, CommandScript};
use crate::error::{ErrorString, PageResult};

use std::path::Path;
use tokio::fs;
use tokio::io::{AsyncBufReadExt, AsyncWriteExt, BufReader};
use tracing::trace;

#[derive(TryFromMultipart)]
struct VpnParam {
    setting_name: String,
    conf_file: FieldData<Bytes>,
    auth_type: String,
    username: String,
    password: String,
    cert_file: FieldData<Bytes>,
    key_file: FieldData<Bytes>,
    key_pass: String,
}

#[derive(Template)]
#[template(path = "../src/vpn/templates/vpn.html")]
struct VpnTemplate {
    setting_name: String,
    vpn_addr: HashMap<&'static str, String>,
}

pub trait VpnRoutes {
    fn vpn_routes(self) -> Self;
}

impl VpnRoutes for Router {
    fn vpn_routes(self) -> Self {
        self.route("/vpn", get(vpn))
            .route("/vpn_setup", post(vpn_setup))
            .route("/vpn_delete", post(vpn_delete))
            .route("/vpn_connect", post(vpn_connect))
            .route("/vpn_disconnect", post(vpn_disconnect))
    }
}

async fn vpn(session: ReadableSession) -> PageResult {
    if !cfg!(debug_assertions) && !session.get::<bool>("login").unwrap_or(false) {
        return Ok(Redirect::to("/").into_response());
    }

    let args = &["vpn_info.sh"];
    let output = common::exec_command(args).await?;
    let setting_name = String::from_utf8_lossy(&output.stdout).trim().to_string();

    let interface = "tun0";
    let vpn_addr = common::current_address(interface, interface).await?;

    let template = VpnTemplate {
        setting_name,
        vpn_addr,
    };
    Ok(HtmlTemplate::new("VPN設定", template).into_response())
}

async fn vpn_setup(
    session: ReadableSession,
    TypedMultipart(vpn_param): TypedMultipart<VpnParam>,
) -> impl IntoResponse {
    if !cfg!(debug_assertions) && !session.get::<bool>("login").unwrap_or(false) {
        return Err(ErrorString {
            message: "not logged in".to_string(),
        });
    }

    // Delete current settings before setup.
    let args = &["vpn_delete.sh"];
    common::exec_command(args).await?;

    let conf_file_contents = vpn_param.conf_file.contents;
    let conf_file_name = vpn_param
        .conf_file
        .metadata
        .file_name
        .context("Config file name is none")?;
    if let Err(e) = save_vpn_file(&conf_file_name, &conf_file_contents).await {
        delete_vpn_dir().await?;
        return Err(anyhow!("Could not save file: {}.", e).into());
    }

    let mut args = vec![
        "vpn_setup.sh".to_string(),
        vpn_param.setting_name,
        vpn_param.auth_type.clone(),
        conf_file_name,
    ];

    if vpn_param.auth_type == "userpass" {
        args.push("--user".to_string());
        args.push(vpn_param.username);
        args.push("--password".to_string());
        args.push(vpn_param.password);
    } else {
        let cert_file_contents = vpn_param.cert_file.contents;
        let cert_file_name = match vpn_param.cert_file.metadata.file_name {
            Some(f) => f,
            None => return Err(anyhow!("Cert file name is none.").into()),
        };
        if !cert_file_contents.is_empty() && !cert_file_name.is_empty() {
            if let Err(e) = save_vpn_file(&cert_file_name, &cert_file_contents).await {
                delete_vpn_dir().await?;
                return Err(anyhow!("Could not save file: {}.", e).into());
            }
            args.push("--cert".to_string());
            args.push(cert_file_name);
        }

        let key_file_contents = vpn_param.key_file.contents;
        let key_file_name = match vpn_param.key_file.metadata.file_name {
            Some(f) => f,
            None => return Err(anyhow!("Key file name is none.").into()),
        };
        if !key_file_contents.is_empty() && !key_file_name.is_empty() {
            if let Err(e) = save_vpn_file(&key_file_name, &key_file_contents).await {
                delete_vpn_dir().await?;
                return Err(anyhow!("Could not save file: {}.", e).into());
            }
            args.push("--key".to_string());
            args.push(key_file_name);
        }
        if !vpn_param.key_pass.is_empty() {
            args.push("--askpass".to_string());
            args.push(vpn_param.key_pass);
        }
    }

    Ok((
        [(CONTENT_TYPE, "text/plain")],
        StreamBody::new(vpn_start(args).await),
    ))
}

async fn vpn_start(args: Vec<String>) -> impl Stream<Item = Result<Vec<u8>>> {
    try_stream! {
    let mut command = command(&args, CommandScript(true), CommandAsRoot(true))?
        .kill_on_drop(true)
        .stdout(Stdio::piped())
        .stderr(Stdio::piped())
        .spawn()
        .context("Could not start vpn")?;
    let mut stdout = BufReader::new(command
        .stdout
        .take()
        .context("vpn_setup.sh command had no stdout")?);
    let mut stdout_buf = vec![];
    let mut stdout_open = true;
    let mut stderr = BufReader::new(command
        .stderr
        .take()
        .context("vpn_setup.sh command had no stderr")?);
    let mut stderr_buf = vec![];
    let mut stderr_open = true;

    while stdout_open || stderr_open {
        tokio::select! {
            n = stdout.read_until(b'\n', &mut stdout_buf), if stdout_open => {
                if n.unwrap_or(0) == 0 && stdout_buf.is_empty() {
                    stdout_open = false;
                } else {
                    trace!("{}", String::from_utf8_lossy(&stdout_buf));
                    if String::from_utf8_lossy(&stdout_buf).contains("Initialization Sequence Completed") {
                        yield stdout_buf;
                        break;
                    }
                    yield stdout_buf;
                    stdout_buf = vec![];
                }
            },
            n = stderr.read_until(b'\n', &mut stderr_buf), if stderr_open => {
                if n.unwrap_or(0) == 0 && stderr_buf.is_empty() {
                    stderr_open = false;
                } else {
                    trace!("{}", String::from_utf8_lossy(&stderr_buf));
                    yield stderr_buf;
                    stderr_buf = vec![];
                }
            },
        }
    }
    let args = &["pkill.sh", "openvpn_log"];
    let _ = common::exec_command(args).await;
    }
}

async fn vpn_delete(session: ReadableSession) -> PageResult {
    if !cfg!(debug_assertions) && !session.get::<bool>("login").unwrap_or(false) {
        return Ok(Redirect::to("/").into_response());
    }

    let args = &["vpn_delete.sh"];
    common::exec_command(args).await?;

    Ok(Redirect::to("/vpn").into_response())
}

async fn vpn_connect(session: ReadableSession) -> PageResult {
    if !cfg!(debug_assertions) && !session.get::<bool>("login").unwrap_or(false) {
        return Ok(Redirect::to("/").into_response());
    }

    let args = &["podman.sh", "start", common::OPENVPN_CONTAINER];
    common::exec_command(args).await?;

    Ok(Redirect::to("/vpn").into_response())
}

async fn vpn_disconnect(session: ReadableSession) -> PageResult {
    if !cfg!(debug_assertions) && !session.get::<bool>("login").unwrap_or(false) {
        return Ok(Redirect::to("/").into_response());
    }

    let args = &["podman.sh", "stop", common::OPENVPN_CONTAINER];
    common::exec_command(args).await?;

    Ok(Redirect::to("/vpn").into_response())
}

async fn save_vpn_file(file_name: &str, data: &[u8]) -> Result<()> {
    let conf_dir = Path::new(common::OPENVPN_CONF_DIR);
    if !conf_dir.exists() {
        fs::create_dir_all(conf_dir).await?;
    }
    let path = conf_dir.join(file_name);
    let mut file = fs::File::create(path).await?;
    file.write_all(data).await?;
    Ok(())
}

async fn delete_vpn_dir() -> Result<()> {
    fs::remove_dir_all(common::OPENVPN_CONF_DIR).await?;
    Ok(())
}
