use anyhow::Context;
use askama::Template;
use axum::{
    extract::Form,
    response::{IntoResponse, Redirect},
    routing::{get, post},
    Router,
};
use axum_sessions::extractors::ReadableSession;
use ipnet::Ipv4Net;
use serde::Deserialize;
use std::{collections::HashMap, net::Ipv4Addr, str::FromStr};

use crate::common;
use crate::common::html_template::HtmlTemplate;
use crate::error::PageResult;

#[derive(Deserialize)]
struct WlanParam {
    essid: String,
    other_essid: String,
    password: String,
    security: String,
    method: String,
    ip_addr: String,
    netmask: String,
    gateway_addr: String,
    p_dns_addr: String,
    s_dns_addr: String,
}

#[derive(Deserialize)]
struct WlanConnectParam {}

#[derive(Deserialize)]
struct WlanDisconnectParam {}

#[derive(Deserialize)]
struct WlanDelParam {}

#[derive(Deserialize, Debug)]
struct APParam {
    bridge_addr: String,
    netmask: String,
    hw_mode: String,
    channel: String,
    ssid: String,
    wpa_passphrase: String,
}

#[derive(Deserialize, Debug)]
struct APDelParam {}

#[derive(Template)]
#[template(path = "../src/wlan/templates/wlan.html")]
struct WlanTemplate<'a> {
    ssid_list: Vec<&'a str>,
    wlan_con_ssid: String,
    security: String,
    current_ip_addr: String,
    current_gw_addr: String,
    current_netmask: String,
    interface: String,
    ap_info: HashMap<&'static str, String>,
    ap_br_interface: String,
    has_wlan: bool,
}

pub trait WlanRoutes {
    fn wlan_routes(self) -> Self;
}

impl WlanRoutes for Router {
    fn wlan_routes(self) -> Self {
        self.route("/wlan", get(wlan))
            .route("/wlan_setup", post(wlan_setup))
            .route("/wlan_delete", post(wlan_delete))
            .route("/wlan_connect", post(wlan_connect))
            .route("/wlan_disconnect", post(wlan_disconnect))
            .route("/ap_setup", post(ap_setup))
            .route("/ap_delete", post(ap_delete))
    }
}

async fn wlan(session: ReadableSession) -> PageResult {
    if !cfg!(debug_assertions) && !session.get::<bool>("login").unwrap_or(false) {
        return Ok(Redirect::to("/").into_response());
    }
    let output = common::exec_command(&["get_ssid_list.sh"]).await?;
    let stdout = String::from_utf8_lossy(&output.stdout);
    let ssid_list = stdout
        .lines()
        .collect::<Vec<&str>>()
        .iter()
        .map(|s| s.trim())
        .collect::<Vec<&str>>();

    let args = &["get_wlan_con_ssid.sh", common::WLAN_CON_NAME];
    let output = common::exec_command(args).await?;
    let stdout = String::from_utf8_lossy(&output.stdout);
    let v = stdout.trim().split(';').collect::<Vec<&str>>();
    let mut ssid = "";
    let mut security = "";
    if v.len() == 2 {
        ssid = if v[0].is_empty() { "--" } else { v[0] };
        security = if v[1].is_empty() { "--" } else { v[1] };
    }

    let interface = common::wlan_interface().unwrap_or("--");
    let current_addr = common::current_address(interface, interface).await?;
    let ap_info = common::ap_info().await?;

    let template = WlanTemplate {
        ssid_list,
        wlan_con_ssid: ssid.to_string(),
        security: security.to_string(),
        current_ip_addr: current_addr["current_ip_addr"].to_owned(),
        current_netmask: current_addr["current_netmask"].to_owned(),
        current_gw_addr: current_addr["current_gw_addr"].to_owned(),
        interface: interface.to_string(),
        ap_info,
        ap_br_interface: common::AP_BRIDGE_INTERFACE.to_string(),
        has_wlan: common::wlan_interface().is_some(),
    };
    Ok(HtmlTemplate::new("WLAN設定", template).into_response())
}

async fn wlan_setup(session: ReadableSession, Form(wlan_param): Form<WlanParam>) -> PageResult {
    if !cfg!(debug_assertions) && !session.get::<bool>("login").unwrap_or(false) {
        return Ok(Redirect::to("/").into_response());
    }

    let interface =
        common::wlan_interface().with_context(|| "Could not find wlan network interface.")?;
    let mut essid = wlan_param.essid;
    if essid == "other" {
        essid = wlan_param.other_essid;
    }
    let mut args = vec![
        "wlan_setup.sh".to_string(),
        common::WLAN_CON_NAME.to_string(),
        essid,
        wlan_param.security,
        wlan_param.password,
        interface.to_string(),
        wlan_param.method.clone(),
    ];

    if wlan_param.method == "manual" {
        let netmask = if wlan_param.netmask.is_empty() {
            "255.255.255.255"
        } else {
            &wlan_param.netmask
        };
        let ip_addr = Ipv4Addr::from_str(&wlan_param.ip_addr)
            .with_context(|| format!("Could not parse addr: {}", &wlan_param.ip_addr))?;
        let netmask = Ipv4Addr::from_str(netmask)
            .with_context(|| format!("Could not parse network: {}", netmask))?;
        let net = Ipv4Net::with_netmask(ip_addr, netmask)
            .with_context(|| format!("Could not define network with {} / {}", ip_addr, netmask))?;

        args.push("-a".to_string());
        args.push(net.to_string());
        if !wlan_param.gateway_addr.is_empty() {
            args.push("-g".to_string());
            args.push(wlan_param.gateway_addr);
        }
        if !wlan_param.p_dns_addr.is_empty() {
            let mut dns = wlan_param.p_dns_addr;
            if !wlan_param.s_dns_addr.is_empty() {
                dns = format!("{},{}", dns, wlan_param.s_dns_addr);
            }
            args.push("-d".to_string());
            args.push(dns);
        }
    }

    common::exec_command(&args).await?;
    Ok(Redirect::to("/wlan").into_response())
}

async fn wlan_delete(session: ReadableSession) -> PageResult {
    if !cfg!(debug_assertions) && !session.get::<bool>("login").unwrap_or(false) {
        return Ok(Redirect::to("/").into_response());
    }

    let args = &["wlan_delete.sh", common::WLAN_CON_NAME];

    common::exec_command(args).await?;

    Ok(Redirect::to("/wlan").into_response())
}

async fn wlan_connect(session: ReadableSession) -> PageResult {
    if !cfg!(debug_assertions) && !session.get::<bool>("login").unwrap_or(false) {
        return Ok(Redirect::to("/").into_response());
    }

    let args = &["/wlan_connect.sh", common::WLAN_CON_NAME, "up"];

    common::exec_command(args).await?;

    Ok(Redirect::to("/wlan").into_response())
}

async fn wlan_disconnect(session: ReadableSession) -> PageResult {
    if !cfg!(debug_assertions) && !session.get::<bool>("login").unwrap_or(false) {
        return Ok(Redirect::to("/").into_response());
    }

    let args = &["wlan_connect.sh", common::WLAN_CON_NAME, "down"];

    common::exec_command(args).await?;

    Ok(Redirect::to("/wlan").into_response())
}

async fn ap_setup(session: ReadableSession, Form(ap_param): Form<APParam>) -> PageResult {
    if !cfg!(debug_assertions) && !session.get::<bool>("login").unwrap_or(false) {
        return Ok(Redirect::to("/").into_response());
    }

    let netmask = if ap_param.netmask.is_empty() {
        "255.255.255.255"
    } else {
        &ap_param.netmask
    };
    let ip_addr = Ipv4Addr::from_str(&ap_param.bridge_addr)
        .with_context(|| format!("Could not parse addr: {}", &ap_param.bridge_addr))?;
    let netmask = Ipv4Addr::from_str(netmask)
        .with_context(|| format!("Could not parse network: {}", netmask))?;
    let net = Ipv4Net::with_netmask(ip_addr, netmask)
        .with_context(|| format!("Could not define network with {} / {}", ip_addr, netmask))?;
    let interface =
        common::ap_interface().with_context(|| "Could not find wlan network interface.")?;

    let args = &[
        "ap_setup.sh",
        common::AP_BRIDGE_NAME,
        &net.to_string(),
        &ap_param.hw_mode,
        &ap_param.channel,
        &ap_param.ssid,
        &ap_param.wpa_passphrase,
        interface,
    ];
    common::exec_command(args).await?;

    Ok(Redirect::to("/wlan").into_response())
}

async fn ap_delete(session: ReadableSession, Form(_): Form<APDelParam>) -> PageResult {
    if !cfg!(debug_assertions) && !session.get::<bool>("login").unwrap_or(false) {
        return Ok(Redirect::to("/").into_response());
    }

    let interface =
        common::ap_interface().with_context(|| "Could not find wlan network interface.")?;

    let args = &["ap_delete.sh", "abos_web_br_ap", interface];
    common::exec_command(args).await?;

    Ok(Redirect::to("/wlan").into_response())
}
